﻿using Google.Api.Gax.ResourceNames;
using Google.Cloud.SecretManager.V1;
using Microsoft.Extensions.Configuration;
using Newtonsoft.Json;

namespace ApiCore.Common
{
    public static class SecretManagerConfigurationExtensions
    {
        /// <summary>
        /// Add SecretManagerConfigurationSource to the build configuration
        /// </summary>
        /// <param name="configurationBuilder"></param>
        public static IConfigurationBuilder AddGoogleSecretsManager<T>(this IConfigurationBuilder configurationBuilder, string credentialFilePath, string secretName)
        {
            if (configurationBuilder == null)
            {
                throw new ArgumentNullException(nameof(configurationBuilder));
            }

            configurationBuilder.Add(new SecretManagerConfigurationSource(credentialFilePath, secretName));

            return configurationBuilder;
        }
    }

    public class SecretManagerConfigurationSource: IConfigurationSource
    {
        private string credentialFilePath;
        private string secretName;

        public SecretManagerConfigurationSource(string credentialFilePath, string secretName)
        {
            this.credentialFilePath = credentialFilePath;
            this.secretName = secretName;
        }

        public IConfigurationProvider Build(IConfigurationBuilder builder)
        {
            return new SecretManagerConfigurationProvider(credentialFilePath, secretName);
        }
    }

    public class SecretManagerConfigurationProvider : ConfigurationProvider
    {
        private readonly SecretManagerServiceClient _client;
        private readonly string? _projectId;
        private string secretName;

        public SecretManagerConfigurationProvider(string credentialFilePath, string secretName)
        {
            this.secretName = secretName;
            var cb = new SecretManagerServiceClientBuilder();
            //cb.CredentialsPath = credentialFilePath;
            string json = File.ReadAllText(credentialFilePath);
            Dictionary<string, string> credData = JsonConvert.DeserializeObject<Dictionary<string, string>>(json) ?? new Dictionary<string, string>();
            cb.JsonCredentials = json;
            _client = cb.Build();
            _projectId = credData["project_id"] ?? "taylorlove-info";
        }

        public SecretManagerConfigurationProvider(SecretManagerServiceClient client, string projectId, string secretName)
        {
            _client = client;
            _projectId = projectId;
            this.secretName = secretName;
        }

        /// <summary>
        /// Load Secrets from Google Secret Manager
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Design", "CA1031:Do not catch general exception types", Justification = "N/A")]
        public override void Load()
        {
            if (string.IsNullOrEmpty(_projectId))
            {
                return; // skip for local debug
            }

            var secret = _client.GetSecret(new SecretName(_projectId, this.secretName));
            var secretVersionName = new SecretVersionName(secret.SecretName.ProjectId, secret.SecretName.SecretId, "latest");
            var secretVersion = _client.AccessSecretVersion(secretVersionName);
            var secretJson = secretVersion.Payload.Data.ToStringUtf8();
            var secretData = JsonConvert.DeserializeObject<Dictionary<string, string>>(secretJson) ?? new Dictionary<string, string>();
            foreach(var kvp in secretData)
            {
                Set("AppSettings:"+ kvp.Key, kvp.Value);
            }
        }

        private static string NormalizeDelimiter(string key)
        {
            return key.Replace("__", ConfigurationPath.KeyDelimiter);
        }
    }

}