﻿using ApiCore.Web.Controllers;
using ApiCore.Web.Models.Attributes;
using Microsoft.AspNetCore.Mvc.ModelBinding.Metadata;
using Microsoft.OpenApi.Any;
using Microsoft.OpenApi.Models;
using Swashbuckle.AspNetCore.Swagger;
using Swashbuckle.AspNetCore.SwaggerGen;
using Swashbuckle.AspNetCore.SwaggerUI;
using System.ComponentModel;
using System.Diagnostics;

namespace ApiCore.Web.AppStart
{
    public class SwaggerConfig
    {
        public static void AddSwaggerGen(SwaggerGenOptions c)
        {
            c.SwaggerDoc("v1", new OpenApiInfo
            {
                Title = "ApiCore",
                Version = "v1",
                //Contact = new OpenApiContact()
                //{
                //    Url = new Uri("https://github.com/Pangamma/PixelStacker"),
                //    Name = "View Source"
                //}
            });

            c.ResolveConflictingActions(apiDesc =>
            {
                return apiDesc.First();
            });

            c.IgnoreObsoleteActions();
            c.IgnoreObsoleteProperties();
            c.ParameterFilter<AcceptableValuesFilter>();
            c.ParameterFilter<DefaultValueFilter>();
            //var defaultSchemaIdSelector = c.SchemaGeneratorOptions.SchemaIdSelector;
            //c.CustomSchemaIds((Type currentClass) =>
            //{
            //    string returnedValue = defaultSchemaIdSelector(currentClass);
            //    if (returnedValue.EndsWith("JsonPayload") && !returnedValue.Equals("JsonPayload"))
            //    {
            //        returnedValue = returnedValue.Replace("JsonPayload", "JP");
            //    }
            //    return returnedValue;
            //});
        }

        public static void UseSwaggerUI(SwaggerUIOptions c)
        {
            c.ShowCommonExtensions();
            c.DocExpansion(DocExpansion.None);
            c.DocumentTitle = "API Core";
            c.EnableTryItOutByDefault();

            // https://github.com/swagger-api/swagger-ui/issues/3832
            c.ConfigObject.AdditionalItems["syntaxHighlight"] = new Dictionary<string, object>
            {
                ["activated"] = false
            };
#if !RELEASE
            c.SwaggerEndpoint(BaseApiController.RoutePrefixWithLeadingSlash+"/swagger/v1/swagger.json", "v1");
            c.RoutePrefix = BaseApiController.RoutePrefixWithTrailingSlash + "swagger";
#endif
        }

        public static void UseSwagger(SwaggerOptions c)
        {
            c.RouteTemplate = BaseApiController.RoutePrefixWithTrailingSlash + "swagger/{documentName}/swagger.json";
            c.PreSerializeFilters.Add((swagger, httpReq) =>
            {
                if (httpReq.Headers.ContainsKey("Host"))
                {
                    //The httpReq.PathBase and httpReq.Headers["X-Forwarded-Prefix"] is what we need to get the base path.
                    //For some reason, they are returning as null/blank. Perhaps this has something to do with how the proxy is configured which I don't have control.
                    //For the time being, the base path is manually set here that corresponds to the APIM API Url Prefix.
                    //In this case we set it to 'sample-app'. 

                    //var basePath = httpReq.Headers["Host"] == "taylorlove.info" ? "" : "";
                    swagger.Servers = new List<OpenApiServer> {
                        // Used on front-end. Do not change this.
                        new OpenApiServer { Url =  $"https://{httpReq.Headers["Host"]}/api" },
                        new OpenApiServer { Url =  $"https://{httpReq.Headers["Host"]}" }
                    };
                }

            });
        }
    }


    public class DefaultValueFilter : IParameterFilter
    {
        public void Apply(OpenApiParameter parameter, ParameterFilterContext context)
        {
            if (context.ApiParameterDescription.ModelMetadata is not DefaultModelMetadata meta) return;
            var propAttrs = meta.Attributes.PropertyAttributes;
            if (propAttrs is null) return;

            foreach (var attr in propAttrs.OfType<DefaultValueAttribute>())
            {
                if (attr.Value != null)
                {
                    parameter.Schema.Example = OpenApiAnyFactory.CreateFromJson(Newtonsoft.Json.JsonConvert.SerializeObject(attr.Value));
                    return;
                }
            }

            var parentModelType = context.ParameterInfo?.ParameterType;
            if (parentModelType != null)
            {
                object? parentModel;
                try
                {
                    parentModel = Activator.CreateInstance(parentModelType);
                }
                catch
                {
                    parentModel = null;
                }

                if (parentModel != null)
                {
                    PropertyDescriptor? prop = TypeDescriptor.GetProperties(parentModel!)[parameter.Name];
                    if (prop is not null)
                    {
                        if (prop.CanResetValue(parentModel)) prop.ResetValue(parentModel);
                        var defVal = prop.GetValue(parentModel);
                        if (defVal != null)
                        {
                            string json = !defVal.GetType().IsEnum
                            ? Newtonsoft.Json.JsonConvert.SerializeObject(defVal)
                            : Newtonsoft.Json.JsonConvert.SerializeObject(defVal.ToString());
                            parameter.Schema.Example = OpenApiAnyFactory.CreateFromJson(json);
                        }
                    }
                }
            }
        }
    }

    public class AcceptableValuesFilter : IParameterFilter
    {
        public void Apply(OpenApiParameter parameter, ParameterFilterContext context)
        {
            if (context.ApiParameterDescription.ModelMetadata is not DefaultModelMetadata meta) return;
            var propAttrs = meta.Attributes.PropertyAttributes;
            if (propAttrs is null) return;

            foreach (var attr in propAttrs.OfType<AcceptableIntValuesAttribute>())
            {
                parameter.Schema.Enum.Clear();
                foreach (var val in attr.AllowableValues)
                {
                    parameter.Schema.Enum.Add(new OpenApiInteger(val));
                }
            }

            foreach (var attr in propAttrs.OfType<AcceptableStringValuesAttribute>())
            {
                parameter.Schema.Enum.Clear();
                foreach (var val in attr.AllowableValues)
                {
                    parameter.Schema.Enum.Add(new OpenApiString(val));
                }
            }
        }
    }

    public class EnumSchemaFilter : ISchemaFilter
    {
        public void Apply(OpenApiSchema model, SchemaFilterContext context)
        {
            if (context.Type.IsEnum)
            {
                model.Enum.Clear();
                Enum.GetNames(context.Type)
                    .ToList()
                    .ForEach(n => model.Enum.Add(new OpenApiString(n)));
            }
        }
    }
}
