﻿using ApiCore.Web.Models;
using ApiCore.Web.Models.Maps;
using ApiCore.Web.Utility;
using GoogleApi.Entities.Maps.Common;
using GoogleApi.Entities.Maps.DistanceMatrix.Request;
using GoogleApi.Entities.Maps.DistanceMatrix.Response;
using GoogleApi.Entities.Places.AutoComplete.Request;
using GoogleApi.Entities.Places.Common;
using IPGeolocation;
using Microsoft.AspNetCore.Mvc;
using System.Diagnostics.CodeAnalysis;
using static GoogleApi.GoogleMaps;

namespace ApiCore.Web.Controllers
{
    public class GeoLocationNode
    {
        [NotNull]
        public double Longitude { get; set; }
        [NotNull]
        public double Latitude { get; set; }
    }

    public class GoogleMapsController : BaseApiController
    {
        private AppSettings config;
        private static LruCache<string, GeoLocationNode> ipCache = new LruCache<string, GeoLocationNode>(5000);

        public GoogleMapsController(AppSettings config)
        {
            this.config = config;
        }

        [HttpGet]
        public async Task<string> GetIp()
        {
            string remoteIp = this.GetRemoteIP() ?? "0.0.0.0";
            string actualRemoteIp = remoteIp;
#if !RELEASE
            if (remoteIp == "127.0.0.1" || remoteIp == "::1")
            {
                using (var client = new HttpClient())
                {
                    var reply = await client.GetAsync("https://api64.ipify.org/?format=text");
                    actualRemoteIp = await reply.Content.ReadAsStringAsync();
                }
            }
#endif
            return actualRemoteIp;
        }

        private string GetRemoteIP()
        {
            var httpClientIp = Request.HttpContext.GetServerVariable("HTTP_CLIENT_IP");
            if (!string.IsNullOrWhiteSpace(httpClientIp))
            {
                return httpClientIp;
            }

            var forward = Request.HttpContext.GetServerVariable("HTTP_X_FORWARDED_FOR");
            if (!string.IsNullOrWhiteSpace(forward))
            {
                return forward;
            }

            var remoteAddr = Request.HttpContext.GetServerVariable("REMOTE_ADDR");
            if (!string.IsNullOrWhiteSpace(remoteAddr))
            {
                return remoteAddr;
            }

            return Request.HttpContext.Connection.RemoteIpAddress?.ToString() ?? "127.0.0.1";
        }

        [HttpGet]
        public async Task<JsonPayload<GeoLocationNode>> GeoLocateByIp(string? ip = null)
        {
            if (ip == "127.0.0.1" || ip == "::1")
                return new JsonPayload<GeoLocationNode>(null, "Cannot use local IP address for geolocation queries.");
            string remoteIp = ip ?? this.GetRemoteIP() ?? "0.0.0.0";


            if (ipCache.TryGetValue(remoteIp, out GeoLocationNode? found))
            {
                return new JsonPayload<GeoLocationNode>(found);
            }

            string actualRemoteIp = remoteIp;
            if (remoteIp == "127.0.0.1" || remoteIp == "::1")
            {
                using (var client = new HttpClient())
                {
                    var reply = await client.GetAsync("https://api64.ipify.org/?format=text");
                    actualRemoteIp = await reply.Content.ReadAsStringAsync();
                }
            }

            if (ipCache.TryGetValue(actualRemoteIp, out GeoLocationNode? found2))
            {
                return new JsonPayload<GeoLocationNode>(found2);
            }

            IPGeolocationAPI api = new IPGeolocationAPI(config.IpGeoLocationApiKey);
            GeolocationParams geoParams = new GeolocationParams();
            geoParams.SetIPAddress(actualRemoteIp);
            geoParams.SetFields("geo,time_zone");
            Dictionary<string, object> geolocationReply = await Task.FromResult(api.GetGeolocation(geoParams));
            if (geolocationReply.TryGetValue("response", out var geo))
            {
                var cast = geo as Geolocation;
                if (cast != null)
                {
                    var node = new GeoLocationNode() { 
                        Latitude = double.Parse(cast.GetLatitude()), 
                        Longitude = double.Parse(cast.GetLongitude()) 
                    };

                    ipCache.Set(actualRemoteIp, node);
#if !RELEASE
                    ipCache.Set(remoteIp, node);
#endif
                    return new JsonPayload<GeoLocationNode>(node);
                }
            }

            return new JsonPayload<GeoLocationNode>(null, "Could not load geolocation.");
        }

        [HttpPost]
        public async Task<JsonPayload<List<Row>>> GetDistances([FromServices] GoogleApi.GoogleMaps.DistanceMatrixApi distanceMatrixApi, DistanceMatrixInput model)
        {
            string ip = await GetIp();
            object output = new { };

            var reply = await distanceMatrixApi.QueryAsync(new DistanceMatrixRequest()
            {
                Key = this.config.GoogleMapsApiKey,
                Channel = "ApiCore",
                Avoid = model.Avoid,
                ArrivalTime = model.ArrivalTime,
                DepartureTime = model.DepartureTime,
                Destinations = model.DestinationPlaceIds.Select(d => new LocationEx(new Place(d))),
                Origins = model.OriginPlaceIds.Select(d => new LocationEx(new Place(d))),
                TrafficModel = model.TrafficModel,
                TransitMode = model.TransitMode,
                TransitRoutingPreference = model.TransitRoutingPreference,
                TravelMode = model.TravelMode,
                Units = model.Units
            });

            return new JsonPayload<List<Row>>(reply.Rows.ToList());
        }

        [HttpGet]
        public async Task<JsonPayload<List<Prediction>>> GetPlaces([FromServices] GoogleApi.GooglePlaces.AutoCompleteApi placesApi,  
            string input)
        {
            double lat = 0;
            double lon = 0;
            
            var geoResponse = await this.GeoLocateByIp(null);
            if (geoResponse != null && geoResponse.IsSuccess)
            {
                lat = geoResponse.Payload?.Latitude ?? 0;
                lon = geoResponse.Payload?.Longitude ?? 0;
            }

            var coords = new GoogleApi.Entities.Common.Coordinate(lat, lon);

            var result = await placesApi.QueryAsync(new PlacesAutoCompleteRequest()
            {
                Input = input,
                //Location = coords,
                LocationBias = new GoogleApi.Entities.Places.Common.LocationBias() { 
                    Location = coords,
                    Radius = 100000,
                    IpBias = false

                },
                Key = config.GoogleMapsApiKey
            });


            return new JsonPayload<List<Prediction>>(result?.Predictions.ToList() ?? new List<Prediction>());
        }
    }
}
