﻿using Microsoft.AspNetCore.Mvc;
using Microsoft.AspNetCore.Mvc.ModelBinding;
using Newtonsoft.Json;
using Newtonsoft.Json.Serialization;
using System.Diagnostics.CodeAnalysis;
using System.Text.Json;

namespace ApiCore.Web.Models
{

    /// <summary>
    /// The intent is to create a standardized JSON payload that can be serialized
    /// and delivered to the front-end. Good for AJAX calls.
    /// </summary>
    public class JsonPayload<T>
    {
        public virtual T? Payload { get; set; }

        [NotNull]
        public virtual bool IsSuccess { get { return !Errors.Any() && StatusCode == ResponseStatusCode.SUCCESS; } }

        [NotNull]
        public virtual List<string> Errors { get; set; } = new List<string>();

        protected string _StatusCode;
        public string StatusCode
        {
            get => _StatusCode;
            set => _StatusCode = ResponseStatusCode.ValueOf(value) ?? ResponseStatusCode.UNKNOWN;
        }

        /// <summary>
        /// Initializes Errors to empty list.
        /// </summary>
        public JsonPayload()
        {
            Errors = new List<string>();
            _StatusCode = ResponseStatusCode.SUCCESS;
        }

        /// <summary>
        /// Successful unless errors are specified later. Empty errors list.
        /// </summary>
        /// <param name="data"></param>
        public JsonPayload(T? data)
        {
            Payload = data;
            Errors = new List<string>();
            _StatusCode = ResponseStatusCode.SUCCESS;
        }

        /// <summary>
        /// Successful unless errors are specified later. Empty errors list.
        /// </summary>
        /// <param name="data"></param>
        /// <param name="errors"></param>
        public JsonPayload(T? data, List<string> errors)
        {
            Payload = data;
            Errors = errors;
            if (errors.Any()) _StatusCode = ResponseStatusCode.NON_SUCCESS;
            else _StatusCode = ResponseStatusCode.SUCCESS;
        }

        public JsonPayload(T? data, string? error)
        {
            Payload = data;
            _StatusCode = ResponseStatusCode.NON_SUCCESS;
            Errors = (error != null) ? new List<string>() { error } : new List<string>();
        }


        /// <summary>
        /// Sets Status to INVALID_MODEL_STATE if any modelstate errors are detected.
        /// </summary>
        /// <param name="model"></param>
        /// <returns></returns>
        public JsonPayload<T> AddAnyErrorsFromModelState(ModelStateDictionary model)
        {
            if (model.IsValid == false)
            {
                var errs = model.Where(x => x.Value?.Errors.Any() ?? false).SelectMany(x => x.Value!.Errors.Select(z => z.ErrorMessage));
                if (errs.Any()) StatusCode = ResponseStatusCode.INVALID_MODEL_STATE;
                Errors.AddRange(errs);
            }

            return this;
        }

        public string ToJsonString()
        {
            return JsonConvert.SerializeObject(this, new JsonSerializerSettings()
            {
                MaxDepth = 10,
                ReferenceLoopHandling = ReferenceLoopHandling.Ignore,
                Formatting = Formatting.Indented,
                NullValueHandling = NullValueHandling.Include,
                ContractResolver = new CamelCasePropertyNamesContractResolver()
            });
        }

        public JsonResult ToJsonResult(int? httpStatusCode = 200)
        {
            try
            {
                var rs = new JsonResult(this, new JsonSerializerOptions()
                {
                    MaxDepth = 10,
                    DefaultIgnoreCondition = System.Text.Json.Serialization.JsonIgnoreCondition.WhenWritingNull,
                    PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
                    WriteIndented = true
                });

                if (httpStatusCode != null)
                {
                    rs.StatusCode = httpStatusCode.Value;
                }

                return rs;
            } 
            catch (InvalidOperationException)
            {
                var rs = new JsonResult(this.ToJsonString(), new JsonSerializerOptions()
                {
                    MaxDepth = 10,
                    DefaultIgnoreCondition = System.Text.Json.Serialization.JsonIgnoreCondition.WhenWritingNull,
                    PropertyNamingPolicy = JsonNamingPolicy.CamelCase,
                    WriteIndented = true
                });

                if (httpStatusCode != null)
                {
                    rs.StatusCode = httpStatusCode.Value;
                }

                return rs;
            }
        }
    }
}
