using ApiCore.Web.AppStart;
using ApiCore.Web.Controllers;
using ApiCore.Web.Models;
using ApiCore.Web.Utility;
using GoogleApi.Extensions;
using Microsoft.AspNetCore.HttpOverrides;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Text.Json;
using System.Text.Json.Serialization;

namespace ApiCore.Web
{
    public class Startup
    {
        public Startup(IConfiguration configuration)
        {
            Configuration = configuration;
            AppSettings.Instance = new AppSettings(configuration);
        }

        public IConfiguration Configuration { get; }

        // This method gets called by the runtime. Use this method to add services to the container.
        public void ConfigureServices(IServiceCollection services)
        {
            services.AddMvc(options =>
            {
                options.EnableEndpointRouting = false;
            })
            .AddJsonOptions(options =>
            {
                options.JsonSerializerOptions.DefaultIgnoreCondition = System.Text.Json.Serialization.JsonIgnoreCondition.WhenWritingNull;
                options.JsonSerializerOptions.PropertyNameCaseInsensitive = true;
                options.JsonSerializerOptions.PropertyNamingPolicy = JsonNamingPolicy.CamelCase;
                options.JsonSerializerOptions.WriteIndented = true;
                options.JsonSerializerOptions.Converters.Add(new JsonStringEnumConverter());
            });

            services.AddSwaggerGen(SwaggerConfig.AddSwaggerGen);
            //services.AddSwaggerGenNewtonsoftSupport();
            services.AddSingleton<AppSettings>(AppSettings.Instance);
            services.AddGoogleApiClients();
        }

        private List<string> GetAllControllerNames()
        {
            var asm = Assembly.GetAssembly(this.GetType());
            var controlleractionlist = asm.GetTypes()
                .Where(type => typeof(Microsoft.AspNetCore.Mvc.Controller).IsAssignableFrom(type))
                .Select(typ => typ.Name)
                .Distinct()
                .ToList();
            return controlleractionlist;
        }

        // This method gets called by the runtime. Use this method to configure the HTTP request pipeline.
        public void Configure(IApplicationBuilder app, IWebHostEnvironment env)
        {
            DeploymentInfoController.ExtraProperties["Environment"] = env;
            // https://stackoverflow.com/a/43878365/1582837
            var forwardedHeadersOptions = new ForwardedHeadersOptions { ForwardedHeaders = ForwardedHeaders.XForwardedFor | ForwardedHeaders.XForwardedProto, RequireHeaderSymmetry = false };
            forwardedHeadersOptions.KnownNetworks.Clear();
            forwardedHeadersOptions.KnownProxies.Clear();

            app
                .UseForwardedHeaders()
                .UseHttpsRedirection()
                .UseMiddleware<CorsHandler>()
                .UseMiddleware<ErrorHandler>()
                .UseStaticFiles()   // Add static files  BEFORE adding routing calls.
            ;


#if API_INCLUDES_DEMO
            string pathFolder = BaseApiController.RoutePrefixWithLeadingSlash;
            // Redirects for the SPA
            {
                // Root index should be the demo
                app.RedirectWhen(path => path == "/", x => $"{pathFolder}/demo/");
                app.RedirectWhen(path => path == "/demo", x => $"{pathFolder}/demo/");
                //var fp = new PhysicalFileProvider(Path.Combine(env.ContentRootPath, "wwwroot", "dist"));
                var fp = new PhysicalFileProvider(Path.Combine(env.ContentRootPath, "PixelStacker.Web.React", "dist"));
                var fsOpts = new FileServerOptions
                {
                    FileProvider = fp,
                    RequestPath = "/demo", // Anything after /demo will try and obtain data from wwwroot/dist.
                    RedirectToAppendTrailingSlash = false,
                };

                fsOpts.EnableDefaultFiles = true;
                fsOpts.DefaultFilesOptions.DefaultFileNames = new string[] { $"index.html" };
                app.UseFileServer(fsOpts);
            }
#else
            // Must redirect to the client side route. Which is /api.
            app.RedirectWhen(path => path == "/", x => $"/api/swagger/index.html");
            //app.RedirectWhen(path => path == "/", x => $"/api/DeploymentInfo/Index");
#endif
            // API Index should be swagger
            List<string> controllerNames = this.GetAllControllerNames().Select(cn => BaseApiController.RoutePrefixWithTrailingSlash + cn.Replace("Contoller", "").ToLowerInvariant()).ToList();

            // Handle API routing
            app.UseWhen((context) => {

                var f = controllerNames.Any(cn => (context.Request.Path.Value?.ToLower().StartsWith(cn) ?? false));
                Debug.WriteLine(String.Join(", ", controllerNames));
                Debug.WriteLine(context.Request.Path.Value?.ToLower() ?? "UNKNOWN");
                return true;
            },
                configWhen => configWhen.UseMvc((Microsoft.AspNetCore.Routing.IRouteBuilder routes) =>
                {
                    routes.MapRoute(
                        name: "specific-api-route",
                        template: BaseApiController.RoutePrefixWithTrailingSlash + "{controller}/{action?}/{id?}",
                        defaults: new { controller = "DeploymentInfo", action = "Index" }
                    );

                    routes.MapRoute(
                        name: "default-api-route",
                        template: "api",
                        defaults: new { controller = "DeploymentInfo", action = "Index" }
                    );
                }
            ));

            app.UseSwagger(SwaggerConfig.UseSwagger);
            app.UseSwaggerUI(SwaggerConfig.UseSwaggerUI);
        }
    }
}
